from typing import Tuple, List


class Schedule:
    def __call__(self, x):
        raise NotImplementedError()


class Flat(Schedule):
    def __init__(self, value):
        self.__value = value

    def __call__(self, x):
        return self.__value

    def __str__(self):
        return f"Schedule({self.__value})"


class Dynamic(Schedule):
    def __init__(self, value):
        self.__value = value

    def __call__(self, x):
        return self.__value

    def update(self, value):
        self.__value = value

    def __str__(self):
        return "Dynamic"


class Piecewise(Schedule):
    """
    ## Piecewise schedule
    """

    def __init__(self, endpoints: List[Tuple[float, float]], outside_value: float = None):
        """
        ### Initialize

        `endpoints` is list of pairs `(x, y)`.
         The values between endpoints are linearly interpolated.
        `y` values outside the range covered by `x` are
        `outside_value`.
        """

        # `(x, y)` pairs should be sorted
        indexes = [e[0] for e in endpoints]
        assert indexes == sorted(indexes)

        self._outside_value = outside_value
        self._endpoints = endpoints

    def __call__(self, x):
        """
        ### Find `y` for given `x`
        """

        # iterate through each segment
        for (x1, y1), (x2, y2) in zip(self._endpoints[:-1], self._endpoints[1:]):
            # interpolate if `x` is within the segment
            if x1 <= x < x2:
                dx = float(x - x1) / (x2 - x1)
                return y1 + dx * (y2 - y1)

        # return outside value otherwise
        return self._outside_value

    def __str__(self):
        endpoints = ", ".join([f"({e[0]}, {e[1]})" for e in self._endpoints])
        return f"Schedule[{endpoints}, {self._outside_value}]"


class RelativePiecewise(Piecewise):
    def __init__(self, relative_endpoits: List[Tuple[float, float]], total_steps: int):
        endpoints = []
        for e in relative_endpoits:
            index = int(total_steps * e[0])
            assert index >= 0
            endpoints.append((index, e[1]))

        super().__init__(endpoints, outside_value=relative_endpoits[-1][1])



from abc import ABC


class StateModule:
    def __init__(self):
        pass

    # def __call__(self):
    #     raise NotImplementedError

    def create_state(self) -> any:
        raise NotImplementedError

    def set_state(self, data: any):
        raise NotImplementedError

    def on_epoch_start(self):
        raise NotImplementedError

    def on_epoch_end(self):
        raise NotImplementedError


class Metric(StateModule, ABC):
    def track(self):
        pass



import dataclasses
from typing import List

import torch

from labml import tracker
from . import Metric


@dataclasses.dataclass
class CollectorState:
    data: List

    def reset(self):
        self.data = []


class Collector(Metric):
    data: CollectorState

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def __call__(self, data: torch.Tensor):
        self.data.data.append(data)

    def create_state(self):
        return CollectorState([])

    def set_state(self, data: any):
        self.data = data

    def on_epoch_start(self):
        self.data.reset()

    def on_epoch_end(self):
        if not self.data.data:
            return
        tracker.add(f"{self.name}.", torch.cat(self.data.data, dim=0))



from typing import Any, TypeVar, Iterator, Iterable, Generic

import torch.nn


class Module(torch.nn.Module):
    r"""
    Wraps ``torch.nn.Module`` to overload ``__call__`` instead of
    ``forward`` for better type checking.
    
    `PyTorch Github issue for clarification <https://github.com/pytorch/pytorch/issues/44605>`_
    """

    def _forward_unimplemented(self, *input: Any) -> None:
        # To stop PyTorch from giving abstract methods warning
        pass

    def __init_subclass__(cls, **kwargs):
        if cls.__dict__.get('__call__', None) is None:
            return

        setattr(cls, 'forward', cls.__dict__['__call__'])
        delattr(cls, '__call__')

    @property
    def device(self):
        params = self.parameters()
        try:
            sample_param = next(params)
            return sample_param.device
        except StopIteration:
            raise RuntimeError(f"Unable to determine"
                               f" device of {self.__class__.__name__}") from None


M = TypeVar('M', bound=torch.nn.Module)
T = TypeVar('T')


class TypedModuleList(torch.nn.ModuleList, Generic[M]):
    def __getitem__(self, idx: int) -> M:
        return super().__getitem__(idx)

    def __setitem__(self, idx: int, module: M) -> None:
        return super().__setitem__(idx, module)

    def __iter__(self) -> Iterator[M]:
        return super().__iter__()

    def __iadd__(self: T, modules: Iterable[M]) -> T:
        return super().__iadd__(modules)

    def insert(self, index: int, module: M) -> None:
        super().insert(index, module)

    def append(self: T, module: M) -> T:
        return super().append(module)

    def extend(self: T, modules: Iterable[M]) -> T:
        return super().extend(modules)

    def forward(self):
        raise NotImplementedError()


if __name__ == '__main__':
    m = Module()
    print(m.device)



import dataclasses

import torch

from labml import tracker
from . import Metric


@dataclasses.dataclass
class AccuracyState:
    samples: int = 0
    correct: int = 0

    def reset(self):
        self.samples = 0
        self.correct = 0


class Accuracy(Metric):
    data: AccuracyState

    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        output = output.view(-1, output.shape[-1])
        target = target.view(-1)
        pred = output.argmax(dim=-1)
        self.data.correct += pred.eq(target).sum().item()
        self.data.samples += len(target)

    def create_state(self):
        return AccuracyState()

    def set_state(self, data: any):
        self.data = data

    def on_epoch_start(self):
        self.data.reset()

    def on_epoch_end(self):
        if self.data.samples == 0:
            return
        tracker.add("accuracy.", self.data.correct / self.data.samples)

    def track(self):
        if self.data.samples == 0:
            return
        tracker.add("accuracy.", self.data.correct / self.data.samples)


class AccuracyDirect(Accuracy):
    data: AccuracyState

    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        output = output.view(-1)
        target = target.view(-1)
        self.data.correct += output.eq(target).sum().item()
        self.data.samples += len(target)



from typing import Generic, TypeVar, Optional

from . import StateModule

T = TypeVar('T')


class SimpleState(Generic[T]):
    state: Optional[T]

    def __init__(self):
        self.state = None

    def get(self) -> T:
        return self.state

    def set(self, data: T):
        self.state = data

    def reset(self):
        self.state = None


class SimpleStateModule(StateModule, Generic[T]):
    data: SimpleState[T]

    def __init__(self):
        super().__init__()

    def set(self, data: T):
        self.data.set(data)

    def get(self) -> T:
        return self.data.get()

    def create_state(self):
        return SimpleState()

    def set_state(self, data: any):
        self.data = data

    def on_epoch_start(self):
        self.data.reset()

    def on_epoch_end(self):
        pass



from typing import Tuple

import torch
from labml import tracker

from labml.configs import BaseConfigs, option, meta_config


class OptimizerConfigs(BaseConfigs):
    optimizer: torch.optim.Adam
    learning_rate: float = 0.01
    momentum: float = 0.5
    parameters: any
    d_model: int
    betas: Tuple[float, float] = (0.9, 0.999)
    eps: float = 1e-08
    step_factor: int = 1024

    def __init__(self):
        super().__init__(_primary='optimizer')


meta_config(OptimizerConfigs.parameters)


@option(OptimizerConfigs.optimizer, 'SGD')
def sgd_optimizer(c: OptimizerConfigs):
    return torch.optim.SGD(c.parameters, c.learning_rate, c.momentum)


@option(OptimizerConfigs.optimizer, 'Adam')
def adam_optimizer(c: OptimizerConfigs):
    return torch.optim.Adam(c.parameters, lr=c.learning_rate,
                            betas=c.betas, eps=c.eps)


class NoamOpt:
    def __init__(self, model_size: int, learning_rate: float, warmup: int, step_factor: int, optimizer):
        self.step_factor = step_factor
        self.optimizer = optimizer
        self.warmup = warmup
        self.learning_rate = learning_rate
        self.model_size = model_size
        self._rate = 0

    def step(self):
        rate = self.rate(tracker.get_global_step() / self.step_factor)
        for p in self.optimizer.param_groups:
            p['lr'] = rate
        self._rate = rate
        self.optimizer.step()

    def rate(self, step):
        factor = self.model_size ** (-0.5) * min(step ** (-0.5), step * self.warmup ** (-1.5))
        return self.learning_rate * factor

    def zero_grad(self):
        self.optimizer.zero_grad()


@option(OptimizerConfigs.optimizer, 'Noam')
def noam_optimizer(c: OptimizerConfigs):
    optimizer = torch.optim.Adam(c.parameters, lr=0.0, betas=c.betas, eps=c.eps)
    return NoamOpt(c.d_model, 1, 2000, c.step_factor, optimizer)


def _test_noam_optimizer():
    import matplotlib.pyplot as plt
    import numpy as np

    opts = [NoamOpt(512, 1, 4000, None),
            NoamOpt(512, 1, 8000, None),
            NoamOpt(2048, 1, 2000, None)]
    plt.plot(np.arange(1, 20000), [[opt.rate(i) for opt in opts] for i in range(1, 20000)])
    plt.legend(["512:4000", "512:8000", "256:4000"])
    plt.title("Optimizer")
    plt.show()


if __name__ == '__main__':
    _test_noam_optimizer()



import pandas as pd
import torch

from typing import Callable, List

from torch.utils.data import TensorDataset


class CsvDataset(TensorDataset):
    data: pd
    y_cols: List
    x_cols: List
    transform: Callable
    test_fraction: float = 0.0
    train: bool

    def __init__(self, file_path: str, y_cols: List, x_cols: List, train: bool = True,
                 transform: Callable = lambda: None, test_fraction: float = 0.0, nrows: int = None):
        self.__dict__.update(**vars())

        self.data = pd.read_csv(**{'filepath_or_buffer': file_path, 'nrows': nrows})

        data_length = len(self.data)

        self.test_size = int(data_length * self.test_fraction)
        self.train_size = data_length - self.test_size

        self.train_data = self.data.iloc[0:self.train_size]
        self.test_data = self.data.iloc[self.train_size:]

        if train:
            x, y = torch.tensor(self.train_data[self.x_cols].values), torch.tensor(self.train_data[self.y_cols].values)
        else:
            x, y = torch.tensor(self.test_data[self.x_cols].values), torch.tensor(self.test_data[self.y_cols].values)

        super(CsvDataset, self).__init__(x, y)
